from . import cli_graphical_plots
from . import cli_graphical_analysis
from . import cli_parametric_images
from . import cli_reference_tissue_models
from . import cli_tac_fitting
from . import cli_tac_interpolation
from . import cli_preproc
from . import cli_plot_tacs
from . import cli_stats

def main():
    print("PETPAL - Positron Emission Tomography Processing and Analysis Library)")


if __name__ == "__main__":
    main()
    