from . import kinetic_modeling
from . import input_function
from . import preproc
from . import utils
from . import visualizations
from . import pipelines
from . import meta
from . import io

def main():
    print("PET-PAL (Positron Emission Tomography Processing and Analysis Library)")
    
if __name__ == "__main__":
    main()
    